import argparse
import os
import shutil
from typing import Optional

import yaml
from deepdiff import DeepDiff  # type: ignore

from ul_py_tool.utils.docker_compose import DockerComposeFile
from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_str2yaml import arg_str2yaml
from ul_py_tool.utils.aseembly import AssemblyFile, AssemblyTarget
from ul_py_tool.utils.colors import FG_GREEN, FG_RED, NC
from ul_py_tool.utils.run_command import run_command
from ul_py_tool.utils.step import Stepper


class CmdReleaseDcf(Cmd):
    release_target: str
    assembly_file: AssemblyFile
    assembly_target: Optional[AssemblyTarget] = None

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--release-target', dest='release_target', type=str, required=True)
        parser.add_argument('--assembly-file', dest='assembly_file', type=arg_str2yaml, required=True, default='assembly.yaml')

    def run(self) -> None:
        stepper = Stepper()
        cwd = os.getcwd()

        with stepper.step(f'{FG_GREEN}check if target exists{NC}'):
            if self.release_target not in self.assembly_file.targets.keys():
                raise argparse.ArgumentTypeError(f'{FG_RED}Target not found in assembly file{NC}')
            self.assembly_target = self.assembly_file.targets[self.release_target]

        with stepper.step(f'{FG_GREEN}check secrets{NC}') as stp:
            for component_name in self.assembly_file.components.keys():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')
                if not os.path.exists(os.path.join(values_dir, 'secret-values.yaml')):
                    stp.add_error(f'{FG_RED}for {component_name} secrets not found in {values_dir}{NC}')
                if not os.path.exists(os.path.join(values_dir, 'values.yaml')):
                    stp.add_error(f'{FG_RED}for {component_name} deploy values not found in {values_dir}{NC}')

        with stepper.step(f'{FG_GREEN}init required repositories{NC}'):
            for component in self.assembly_file.components.values():
                os.makedirs(os.path.join(cwd, component.directory), exist_ok=False)
                run_command([f'git clone  --depth 1 --branch {component.tag} {component.repository} .'], cwd=os.path.join(cwd, component.directory))
                run_command(['git submodule sync --recursive && git submodule update --init --recursive'], cwd=os.path.join(cwd, component.directory))

        with stepper.step(f'{FG_GREEN}check values and secrets data{NC}') as stp:
            ''' format of DeepDiff output looks like:
            {
              'dictionary_item_added': [root['broker'], root['db'], root['dockerconfig'], root['systemconfig'], root['pii_keys']],
              'dictionary_item_removed': [root['endpoints']]
            }
            or just {} if no diff in keys
            where dictionary_item_added - keys that are not presented  in first dict(yaml)(dev-values/dev-secret-values.yaml)
            and dictionary_item_removed - keys that are presented in first dict(yaml)(dev-values/dev-secret-values.yaml) but missing in second (target)
            '''
            for component_name, component_config in self.assembly_file.components.items():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')

                with open(os.path.join(cwd, component_config.directory, '.helm/dev-values.yaml'), 'r') as f:
                    dev_v = yaml.safe_load(f)

                with open(os.path.join(values_dir, "values.yaml"), 'r') as f:
                    target_v = yaml.safe_load(f)
                ddiff = DeepDiff(dev_v, target_v, ignore_order=True)
                ddiff.pop('values_changed', None)
                if ddiff:
                    stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in values:{NC}\n {ddiff}')

                with open(os.path.join(cwd, component_config.directory, '.helm/dev-secret-values.yaml'), 'r') as f:
                    dev_secret_v = yaml.safe_load(f)

                with open(os.path.join(values_dir, "secret-values.yaml"), 'r') as f:
                    target_secret_v = yaml.safe_load(f)
                ddiff = DeepDiff(dev_secret_v, target_secret_v, ignore_order=True)
                ddiff.pop('values_changed', None)
                if ddiff:
                    stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in secret values:{NC}\n {ddiff}')

        with stepper.step(f'{FG_GREEN}release{NC}'):
            # to store releases that completed
            for component_name, component_config in self.assembly_file.components.items():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')
                component_directory = os.path.join(cwd, component_config.directory)
                release_name = f'{component_name}-{self.assembly_target.environment}'
                # download helm chart
                run_command(
                    [
                        f'helm pull {component_config.helm_repo_name}/{component_config.helm_chart_name} --untar --version {component_config.helm_chart_version}',
                    ],
                    cwd=component_directory,
                )
                if not os.path.exists(os.path.join(cwd, component_config.directory, '.helm/templates')):
                    shutil.copytree(
                        os.path.join(cwd, component_config.directory, f"{component_config.helm_chart_name}/templates"),
                        os.path.join(cwd, component_config.directory, '.helm/templates'),
                    )
                shutil.copy(os.path.join(values_dir, "secret-values.yaml"), component_directory)
                shutil.copy(os.path.join(values_dir, "values.yaml"), component_directory)
                shutil.copy(os.path.join(cwd, component_config.directory, '.helm/charts.yaml'), component_directory)
                enc_key = os.environ[self.assembly_target.encryption_key_name]
                run_command(
                    [
                        f'git add . '
                        f'&& export RELEASE_TAG={component_config.tag} '
                        f'&& export WERF_SECRET_KEY={enc_key} '
                        f'&& export WERF_ENV={self.assembly_target.environment} '
                        f'&& werf render --config=werf-release.yaml  --namespace={release_name} --values=values.yaml  --values=charts.yaml --secret-values=secret-values.yaml --loose-giterminism=True > kubernetes-objects.yaml',
                    ],
                    cwd=component_directory,
                    silent=True,
                )
                with open(os.path.join(component_directory, "kubernetes-objects.yaml"), "r") as fr:
                    dcf = DockerComposeFile.from_kubernetes(yaml.load_all(fr, yaml.FullLoader))
                final_file = dcf.to_json()
                with open(f"docker-compose.{component_name}.release.yaml", "w") as fw:
                    yaml.dump(final_file, fw)



import os.path
import sys

if __name__ == "__main__":
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from ul_py_tool.commands.cmd import Cmd


def main() -> None:
    Cmd.main({
        'code_stats': 'ul_py_tool.commands.cmd_stats:CmdStats',
        'minor': 'ul_py_tool.commands.cmd_version:CmdVersionMinor',
        'major': 'ul_py_tool.commands.cmd_version:CmdVersionMajor',
        'patch': 'ul_py_tool.commands.cmd_version:CmdVersionPatch',
        'lint': 'ul_py_tool.commands.cmd_lint:CmdLint',
        'install': 'ul_py_tool.commands.cmd_install:CmdInstall',
        'test': 'ul_py_tool.commands.cmd_test:CmdTest',
        'fmt': 'ul_py_tool.commands.cmd_fmt:CmdFmt',
        'release': 'ul_py_tool.commands.cmd_release:CmdRelease',
        'release-dcf': 'ul_py_tool.commands.cmd_release_dcf:CmdReleaseDcf',
        'test_secrets': 'ul_py_tool.commands.cmd_test_secrets:CmdTestSecrets',
        'build-images': 'ul_py_tool.commands.cmd_build_images:CmdBuildImages',
        'outdated': 'ul_py_tool.commands.cmp_outdated:CmdOutdated',
        'cleanup': 'ul_py_tool.commands.cmd_cleanup:CmdCleanup',
        'fix_own': 'ul_py_tool.commands.cmd_fix_own:CmdFixOwn',
    })


if __name__ == '__main__':
    main()



from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.step import Stepper


class CmdFixOwn(Cmd):

    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('fix') as stp:
            stp.run_cmd('sudo chown $(whoami):$(whoami) -R .')



import os.path
from argparse import ArgumentTypeError


def arg_file_exists(value: str) -> str:
    if os.path.exists(value):
        return value
    new_val = os.path.join(os.getcwd(), value)
    if os.path.exists(new_val):
        return new_val
    raise ArgumentTypeError(f'file "{value}" was not found')



import os

MAX_LINE_LEN = 180
INDENT_SIZE = 4

THIS_LIB_PATH = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

PYENV_PY_VERSION = '.python-version'

MYPY_CONFIG = 'mypy.ini'

PY_TYPED = 'py.typed'

PIPFILE = 'Pipfile'
PIPFILE_LOCK = 'Pipfile.lock'

SOURCE_PYPI = 'pypi'
SOURCE_NERO_GITLAB = 'nero_gitlab'

GITLAB_TOKEN = 'kZG7esv72n325czD7Wdg'

SOURCE_NERO_GITLAB_HOST = 'gitlab.neroelectronics.by'
SOURCE_NERO_GITLAB_URL = f'https://__token__:{GITLAB_TOKEN}@gitlab.neroelectronics.by/api/v4/projects/996/packages/pypi/simple'

# curl --header "PRIVATE-TOKEN: kZG7esv72n325czD7Wdg" "https://gitlab.neroelectronics.by/api/v4/projects/527/registry/repositories"

HELM_ERROR__NOT_FOUND = 'No results found'
KUEBERNETES_SLEEP_TIME = 60

CHANGE_LOG = 'CHANGELOG.md'

SUPPORTED_PYTHON_VERSIONS = {'3.8', '3.10'}

PRE_PUSH_DEST = os.path.join(os.getcwd(), ".git", "hooks", "pre-push")
PRE_COMMIT_DEST = os.path.join(os.getcwd(), ".git", "hooks", "pre-commit")

UNDER_DOCKER = os.path.exists('/docker_app')

UNDER_CI_JOB = bool(os.environ.get('CI_JOB_ID')) and bool(os.environ.get('CI_COMMIT_SHA')) and bool(os.environ.get('CI_REPOSITORY_URL'))



import argparse
import glob
import os
from typing import List, Callable, Optional, Union, Tuple

from ul_py_tool.utils.write_stdout import write_stdout


def arg_file_info(file: str, *, rel_path: bool = True, ide_col_suf: bool = True, line: int = 0, col: Optional[int] = None) -> str:
    suf = f':{line}{f":{col}" if col is not None else ""}' if ide_col_suf else ''
    return f'./{os.path.relpath(file, os.getcwd())}{suf}' if rel_path else file


def arg_files_print(
    limit: int,
    files: List[str],
    *,
    ignored_files: Optional[List[str]] = None,
    name: str = '',
    ide_col_suf: bool = True,
    rel_path: bool = True,
    print_total: bool = True,
) -> None:
    if limit == 0:
        return
    if limit < 0:
        limit = len(files)
    if name:
        write_stdout(f'{name}:')
    for f in files[:limit]:
        write_stdout(arg_file_info(f, ide_col_suf=ide_col_suf, rel_path=rel_path))

    hidden_str = '' if len(files) <= limit else f'... {len(files) - limit} hidden. '
    ignored_str = '' if ignored_files is None else f'{len(ignored_files)} files ignored. '

    if print_total:
        write_stdout(f'... {len(files)} files found. {hidden_str}{ignored_str}')


def arg_file_glob_compile_files(include: Union[List[str], List[List[str]]], exclude: Optional[Union[List[str], List[List[str]]]] = None) -> Tuple[List[str], List[str]]:
    include_files = set()
    for inc in include:
        if not isinstance(inc, (list, tuple)):
            inc = [inc]
        for f in inc:
            assert isinstance(f, str), f'path must be str. "{type(f).__name__}" was given'
            include_files.add(os.path.abspath(f))

    exclude_files = set()
    for exc in (exclude or []):
        if not isinstance(exc, (list, tuple)):
            exc = [exc]
        for f in exc:
            assert isinstance(f, str), f'path must be str. "{type(f).__name__}" was given'
            exclude_files.add(os.path.abspath(f))

    res_files = include_files - exclude_files

    return list(sorted(res_files)), list(sorted(include_files - res_files))


def arg_files_glob(*, ignore_absent: bool = False, dir: bool = False) -> Callable[[str], List[str]]:
    def arg_files_glob_wr(value: str) -> List[str]:
        files: List[str] = list()
        if '*' in value:
            for gf in glob.iglob(value, recursive=True):
                gf_s = os.path.abspath(str(gf))
                if (dir and os.path.isdir(gf_s)) or (not dir and os.path.isfile(gf_s)):
                    files.append(os.path.abspath(gf_s))
        else:
            if (dir and os.path.isdir(value)) or (not dir and os.path.isfile(value)):
                files.append(os.path.abspath(value))
        if not ignore_absent and len(files) == 0:
            raise argparse.ArgumentTypeError(f'no files found with path template "{value}"')
        return files

    return arg_files_glob_wr



import argparse
import os
from glob import glob
from typing import Set, List
from urllib.parse import urlparse, ParseResult

import gitlab
import requests

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_str2bool import arg_str2bool
from ul_py_tool.utils.arg_str2list import arg_str2list
from ul_py_tool.utils.docker_file import DockerFile
from ul_py_tool.utils.run_command import run_command
from ul_py_tool.utils.step import Stepper


class CmdBuildImages(Cmd):
    silent: bool
    push: bool
    registry: str
    images_dir: str
    project_id: int
    api_token: str
    image_version: str
    build_dir: str
    build_args: List[str]
    gitlab: bool

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--registry', required=False, default="", type=str)
        parser.add_argument('--images-dir', required=False, default="", type=str)
        parser.add_argument('--silent', required=False, default=False, type=arg_str2bool)
        parser.add_argument('--push', required=False, default=False, type=arg_str2bool)
        parser.add_argument('--project_id', required=False, default="", type=int)
        parser.add_argument('--api-token', required=False, default="", type=str)
        parser.add_argument('--image-version', required=False, default="", type=str)
        parser.add_argument('--build-dir', required=False, default="", type=str)
        parser.add_argument('--build-args', required=False, default=[], type=arg_str2list)
        parser.add_argument('--gitlab', required=False, default=True, type=arg_str2bool)

    @staticmethod
    def _remove_prefix(text: str, prefix: str) -> str:
        if prefix and text.startswith(prefix):
            return text[len(prefix):]
        return text

    @staticmethod
    def _remove_suffix(text: str, suffix: str) -> str:
        if suffix and text.endswith(suffix):
            return text[:-len(suffix)]
        return text

    @staticmethod
    def _remove_port(text: str) -> str:
        if ':' in text:
            text_without_port = text.split(':')
            return text_without_port[0]
        return text

    @staticmethod
    def _get_gitlab_repository_urls(url: str, project_id: int, registry_url_path: str, registry_url_parsed: ParseResult, token: str) -> Set[str]:
        gl = gitlab.Gitlab(url, private_token=token)
        gl.auth()
        project = gl.projects.get(project_id)
        image_repositories = project.repositories.list()
        return {f'{registry_url_parsed.netloc}/{registry_url_path}/{image_repository.name}:{tag.name}'
                for image_repository in image_repositories for tag in image_repository.tags.list()}

    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('prepare variables for build and push'):
            registry_url_parsed = urlparse(self.registry)
            if len(str(registry_url_parsed.scheme)):
                registry_url_parsed = urlparse(f'https://{self.registry}')

            registry_url_path = self._remove_suffix(self._remove_prefix(str(registry_url_parsed.path), "/"), "/")

            url = f'https://{self._remove_port(registry_url_parsed.netloc)}'
            if self.gitlab:
                image_repository_urls = self._get_gitlab_repository_urls(url, self.project_id, registry_url_path, registry_url_parsed, self.api_token)
            else:
                image_repository_urls = set()
                repositories = requests.get(
                    f"https://hub.docker.com/v2/repositories/{registry_url_path}/?page_size=1000").json()
                for r in repositories['results']:
                    tags = requests.get(
                        f"https://hub.docker.com/v2/repositories/{registry_url_path}/{r['name']}/tags/?page_size=1000").json()
                    for t in tags['results']:
                        image_repository_urls.add(f"{registry_url_path}/{r['name']}:{t['name']}")

        with stepper.step('build and push target images'):
            if not self.gitlab:
                run_command([f"docker login -u {registry_url_path} -p {self.api_token}"], silent=True)
            for image_dir in glob(os.path.join(self.images_dir, "*"), recursive=False):
                if not os.path.isdir(image_dir):
                    continue

                image_version_file = os.path.join(image_dir, "version.txt")
                image_file = os.path.join(image_dir, "Dockerfile")
                image_name = os.path.basename(image_dir)
                if not self.image_version:
                    with open(image_version_file, "rt") as f:
                        self.image_version = "".join(f.readlines()).strip()

                if self.gitlab:
                    image_url = f'{registry_url_parsed.netloc}/{registry_url_path}/{image_name}:{self.image_version}'
                else:
                    image_url = f'{registry_url_path}/{image_name}:{self.image_version}'

                if image_url in image_repository_urls:
                    continue
                else:
                    df = DockerFile.load_file(image_file, image_name)
                    df.build(image_url, self.build_args, self.build_dir).display(disable=self.silent)

                    if self.push:
                        df.push(image_url)
